
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <mem.h>
#include <string.h>

#include "cd.h"
#include "dpmi.h"
#include "cgdk.h"
#include "timer.h"

#define LOCK 1
#define UNLOCK 0

#define CDBUF_SIZE 512

#define CMD_EJECT_TRAY 0
#define CMD_CLOSE_TRAY 5
#define CMD_RESET 2

#define BUSY  512
#define TRACK_MASK 208


/* cd_info.ERROR Description

  Bit 15         - Error bit
  Bit 14-10      - Reserved
  Bit  9         - Busy
  Bit  8         - Done
  Bit  7-0       - Error code (bit 15 on)

Error codes are the following:

  0  Write-protect violation
  1  Unknown unit
  2  Drive not ready
  3  Unknown command
  4  CRC error
  5  Bad drive request structure length
  6  Seek error
  7  Unknown media
  8  Sector not found
  9  Printer out of paper
  A  Write fault
  B  Read fault
  C  General failure
  D  Reserved
  E  Reserved
  F  Invalid disk change
*/


/*   Multiplex Interrupt routines
     "borrowed" from Ralf Brown's MSDOS Interrupt List v4.1

short 21 - CD-ROM device driver - IOCTL INPUT
        AX = 4402h
        BX = file handle referencing character device for CD-ROM driver
        CX = number of bytes to read
        DS:DX -> control block (see #0563)
Return: CF clear if successful
            AX = number of bytes actually read
        CF set on error
            AX = error code (01h,05h,06h,0Dh) (see #0770 at AH=59h)
Note:   the data returned depends on the first byte of the control block; the
          remainder of the control block is filled by the driver
SeeAlso: AX=4403h"CD-ROM",short 2F/AX=0802h

(Table 0562)
Values for CD-ROM data being requested:
 00h    device driver header address
 01h    drive head location
 02h    reserved
 03h    error statistics
 04h    audio channel info
 05h    raw drive bytes (uninterpreted and device-specific)
 06h    device status
 07h    sector size
 08h    volume size
 09h    media change status
 0Ah    audio disk info
 0Bh    audio track info
 0Ch    audio Q-Channel info
 0Dh    audio sub-channel info
 0Eh    UPC code

Format of CD-ROM control block:
Offset  Size    Description     (Table 0563)
 00h    BYTE    data being requested (see #0562)
---function 00h---
 01h    DWORD   device driver header address (see also AH=52h)
---function 01h---
 01h    BYTE    addressing mode
                00h HSG
                01h Red Book
 02h    DWORD   current location of drive's head
                logical sector number in HSG mode
                frame/second/minute/unused in Red Book mode
                (HSG sector = minute * 4500 + second * 75 + frame - 150)
---function 03h---
 01h  N BYTEs   undefined as of 5 Aug 88 specification
---function 04h---
 01h    BYTE    input channel (0-3) for output channel 0
 02h    BYTE    volume for output channel 0
 03h    BYTE    input channel (0-3) for output channel 1
 04h    BYTE    volume for output channel 1
 05h    BYTE    input channel (0-3) for output channel 2
 06h    BYTE    volume for output channel 2
 07h    BYTE    input channel (0-3) for output channel 3
 08h    BYTE    volume for output channel 3
Notes:  output channels 0 and 1 are left and right, 2 and 3 are left prime and
          right prime; a volume of 00h is off
        the default setting is for each input channel to be assigned to the
          same-numbered output channel at full (FFh) volume
---function 05h---
 01h    BYTE    number of bytes read
 02h 128 BYTEs  buffer for drive bytes
---function 06h---
 01h    DWORD   device parameters (see #0564)
---function 07h---
 01h    BYTE    read mode
                00h cooked
                01h raw
 02h    WORD    sector size in bytes
---function 08h---
 01h    DWORD   volume size in sectors
---function 09h---
 01h    BYTE    media change status
                00h don't know
                01h media unchanged
                FFh media has been changed
---function 0Ah---
 01h    BYTE    lowest audio track number
 02h    BYTE    highest audio track number
 03h    DWORD   start address of lead-out track (Red Book format)
--function 0Bh---
 01h    BYTE    track number (set by caller)
 02h    DWORD   starting point of track (Red Book format)
 06h    BYTE    track control info
                bits 15,14,12: track type (notice: bits not contiguous!)
                        000 two audio channels, no pre-emphasis
                        001 two audio channels with pre-emphasis
                        010 data track
                        100 four audio channels, no pre-emphasis
                        101 four audio channels with pre-emphasis
			other reserved
		bit 13: digital copy permitted
---function 0Ch---
 01h    BYTE    CONTROL and ADR byte (as received from drive)
 02h    BYTE    track number
 03h    BYTE    point or index
 04h    BYTE    minute  \
 05h    BYTE    second   > running time within track
 06h    BYTE    frame   /
 07h    BYTE    zero
 08h    BYTE    "AMIN" or "PMIN"     \
 09h    BYTE    "ASEC" or "PSEC"      > running time on disk
 0Ah    BYTE    "AFRAME" or "PFRAME" /
---function 0Dh---
 01h    DWORD   starting frame address (Red Book format)
 05h    DWORD   transfer address
 09h    DWORD   number of sectors to read
Note:   copies 96 bytes of sub-channel info per sector into buffer
---function 0Eh---
 01h    BYTE    CONTROL and ADR byte
 02h  7 BYTEs   UPC/EAN code (13 BCD digits,low-order nybble of last byte is 0)
 09h    BYTE    zero
 0Ah    BYTE    "AFRAME"

Bitfields for CD-ROM device parameters:
Bit(s)  Description     (Table 0564)
 0      door open
 1      door unlocked
 2      supports raw reading in addition to cooked
 3      writable
 4      can play audio/video tracks
 5      supports interleaving
 6      reserved
 7      supports prefetch requests
 8      supports audio channel control
 9      supports Red Book addressing in addition to HSG
 10     audio is playing
*/

int CD_installed;
TCD_data CD_info;
TCD_playInfo CD_playInfo;

void *CDbuf, *CDbuf2;
short CDbuf2_low, CDbuf2_high;

typedef struct
{
  char code;
  short off;
  short seg;
} TCD_driveList;

char CD_driverName [8+1];
int CD_handle;

static int CD_door;
#define OPENED 1
#define CLOSED 2


static void allocbuffers (void)
{
  if (CDbuf == NULL)
    CDbuf = DPMI_alloc (CDBUF_SIZE);
  if (CDbuf == NULL)
    CGDK_exit ("Not enough memory for 16-bit CD-ROM buffer.\n", EXIT_FAILURE);
  if (CDbuf2 == NULL)
    CDbuf2 = DPMI_alloc (CDBUF_SIZE);
  if (CDbuf2 == NULL)
    CGDK_exit ("Not enough memory for 16-bit CD-ROM buffer.\n", EXIT_FAILURE);
  CDbuf2_low = DP_OFF (CDbuf2);
  CDbuf2_high = DP_SEG (CDbuf2);
}


static void device_request ()
{
  RMINFO rmi={0};

  rmi.x.eax = 0x00001510;
  rmi.x.ecx = CD_info.firstDrive;
  rmi.x.ebx = DP_OFF (CDbuf);
  rmi.x.es = DP_SEG (CDbuf);

  DPMI_int (0x2F, &rmi);
/*  if (outregs.x.cflag)
    printf ("DEVICE REQUEST FAILED!!!\n");*/
}


static void red_book (TCD_time *time, unsigned long value)
{
  time->frame = value & 0x000000ff;
  time->sec = (value & 0x0000ff00) >> 8;
  time->min = (value & 0x00ff0000) >> 16;
}

static unsigned long hsg (TCD_time *time)
{
  unsigned long value;

  value = (unsigned long)time->min * 4500;
  value += (short)time->sec * 75;
  value += time->frame;
//  value -= 150;
  return value;
}


unsigned long CD_headPosition (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
    unsigned char unused[4];
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char adr_mode;
    unsigned long address;
  } head_data;

  tray_request.length = sizeof (tray_request);
  tray_request.subunit = 0;
  tray_request.comcode = 3;
  tray_request.media = tray_request.sector = tray_request.volid = 0;
  tray_request.address_low = CDbuf2_low;
  tray_request.address_high = CDbuf2_high;
  tray_request.bytes = 6;
  head_data.mode = 0x01;
  head_data.adr_mode = 0x00;
  memcpy (CDbuf, &tray_request, tray_request.length);
  memcpy (CDbuf2, &head_data, 6);
  device_request ();
  memcpy (&tray_request, CDbuf, tray_request.length);
  memcpy (&head_data, CDbuf2, 6);
  memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  return head_data.address;
}


void CD_getVolume (TCD_volumeInfo *vol)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 3;
    tray_request.media = 0;
    tray_request.media = tray_request.sector = tray_request.volid = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 9;
    vol->mode = 0x04;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, vol, sizeof (TCD_volumeInfo));
    device_request ();
    memcpy (&tray_request, CDbuf, tray_request.length);
    memcpy (vol, CDbuf2, sizeof (TCD_volumeInfo));
    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  }
}


void CD_setVolume (TCD_volumeInfo *vol)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned char unused[4];
  } CD_request;

  if (CD_installed == CD_INSTALLED)
  {
    vol->mode = 3;
    CD_request.length = sizeof (CD_request);
    CD_request.subunit = 0;
    CD_request.comcode = 12;
    CD_request.media = 0;
    CD_request.address_low = CDbuf2_low;
    CD_request.address_high = CDbuf2_high;

    CD_request.bytes = 9;
    memcpy (CDbuf, &CD_request, sizeof (CD_request));
    memcpy (CDbuf2, vol, sizeof (TCD_volumeInfo));
    device_request ();
    memcpy (&CD_request, CDbuf, CD_request.length);
    memcpy (&CD_info.error, &CD_request.status, sizeof (short));
  }
}


void CD_getUPC (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char adr;
    unsigned char upc[7];
    unsigned char zero;
    unsigned char aframe;
  } upc_data;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 3;
    tray_request.media = 0;
    tray_request.media = tray_request.sector = tray_request.volid = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 11;
    upc_data.mode = 0x0e;
    upc_data.adr = 2;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &upc_data, 11);
    device_request ();
    memcpy (&upc_data, CDbuf2, 11);
    memcpy (&tray_request, CDbuf, tray_request.length);
    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
    if (upc_data.adr == 0)
      memset (&upc_data.upc, 0, 7);
    memcpy (&CD_info.upc[0], &upc_data.upc[0], 7);
  }
}


void CD_audioInit (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    short address_low;
    short address_high;
    short bytes;
    short sector;
    long volid;
  } ioctli;

  struct {
    unsigned char mode;
    unsigned char lowest;
    unsigned char highest;
    unsigned long address;
  } track_data;

  if (CD_installed == CD_INSTALLED)
  {
    ioctli.length = sizeof (ioctli);
    ioctli.subunit = 0;
    ioctli.comcode = 3;
    ioctli.media = 0;
    ioctli.sector = 0;
    ioctli.volid = 0;
    ioctli.address_low = CDbuf2_low;
    ioctli.address_high = CDbuf2_high;

    ioctli.bytes = 7;
    memset (&track_data, 0, 7);
    track_data.mode = 0x0a;
    memcpy (CDbuf, &ioctli, sizeof (ioctli));
    memcpy (CDbuf2, &track_data.mode, 7);
    memcpy (&track_data, CDbuf2, sizeof (track_data));
    device_request ();
    memcpy (&track_data, CDbuf2, sizeof (track_data));
    memcpy (&ioctli, CDbuf, sizeof (ioctli));
    memcpy (&CD_info.diskid, &track_data.lowest, 6);

    CD_info.lowAudio = track_data.lowest;
    CD_info.highAudio = track_data.highest;
    CD_info.tracksTotal = CD_info.highAudio - CD_info.lowAudio + 1;

    red_book (&CD_info.diskLength,track_data.address);
    memcpy (&CD_info.error, &ioctli.status, sizeof (short));

    CD_setTrack (0);//
    CD_info.audioStartTime = CD_info.trackBegin;

    CD_getStatus ();
    CD_getUPC ();
  }
}


static void switch_track (int tracknum)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char track;
    unsigned long address;
    unsigned char control;
  } track_data;

  tray_request.length = sizeof (tray_request);
  tray_request.subunit = 0;
  tray_request.comcode = 3;
  tray_request.media = 0;
  tray_request.media = tray_request.sector = tray_request.volid = 0;
  tray_request.address_low = CDbuf2_low;
  tray_request.address_high = CDbuf2_high;

  tray_request.bytes = 7;
  track_data.mode = 0x0b;
  track_data.track = tracknum+CD_info.lowAudio;
  memcpy (CDbuf, &tray_request, sizeof (tray_request));
  memcpy (CDbuf2, &track_data, sizeof (track_data));
  device_request ();
  memcpy (&tray_request, CDbuf, sizeof (tray_request));
  memcpy (&track_data, CDbuf2, sizeof (track_data));

  memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  red_book (&CD_info.trackBegin, track_data.address);
  CD_info.trackType = track_data.control & TRACK_MASK;
}


static TCD_time operator - (TCD_time &finish, TCD_time &start)
{
  TCD_time newtime = finish;

  if (newtime.frame >= start.frame)
    newtime.frame -= start.frame;
  else
  {
    newtime.frame -= start.frame-100;
    if (newtime.sec == 0)
    {
      newtime.min--;
      newtime.sec=60;
    }
    newtime.sec--;
  }

  if (newtime.sec >= start.sec)
    newtime.sec -= start.sec;
  else
  {
    newtime.sec -= start.sec-60;
    newtime.min --;
  }

  newtime.min -= start.min;
  return newtime;
}


void CD_setTrack (int tracknum)
{
  TCD_time finish;

  if (CD_installed == CD_INSTALLED)
  {
    if (tracknum < CD_info.tracksTotal-1)
    {
      switch_track (tracknum+1);
      finish = CD_info.trackBegin;
      switch_track (tracknum);
    }
    else
    {
      switch_track (tracknum);
      finish = CD_info.diskLength-CD_info.audioStartTime;
    }

    CD_info.trackLength = finish-CD_info.trackBegin;

    CD_info.trackCur = tracknum;

//    CD_getUPC ();
  }
}



void CD_getStatus (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  struct {
    unsigned char mode;
    unsigned long status;
  } CD_data;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 3;
    tray_request.media = 0;
    tray_request.media = tray_request.sector = tray_request.volid = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 5;

    CD_data.mode = 0x06;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &CD_data, sizeof (CD_data));
    device_request ();
    memcpy (&tray_request, CDbuf, sizeof (tray_request));
    memcpy (&CD_data, CDbuf2, sizeof (CD_data));

    memcpy (&CD_info.status, &CD_data.status, sizeof (long));
    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  }
}


void CD_seek (unsigned long location)
{
  unsigned char min, sec, frame;
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char addressmode;
    unsigned long transfer;
    unsigned short sectors;
    unsigned long seekpos;
  } play_request;

  if (CD_installed == CD_INSTALLED)
  {
    play_request.length = sizeof (play_request);
    play_request.subunit = 0;
    play_request.comcode = 131;
    play_request.addressmode = 0;
    play_request.transfer = 0;
    play_request.sectors = 0;
    play_request.seekpos = location;
    memcpy (CDbuf, &play_request, sizeof (play_request));
    device_request ();
    memcpy (&play_request, CDbuf, sizeof (play_request));
    memcpy (&CD_info.error, &play_request.status, sizeof (short));
  }
}


void CD_play (TCD_time *begin, TCD_time *length)
{
  unsigned long leng;

  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char addressmode;
    unsigned long start;
    unsigned long playlength;
  } play_request;

  if (CD_installed == CD_INSTALLED)
  {
    play_request.length = sizeof (play_request);
    play_request.subunit = 0;
    play_request.comcode = 132;
    play_request.addressmode = 0;
    play_request.start = hsg (begin) - hsg (&CD_info.audioStartTime);
    play_request.playlength = hsg (length);
    memcpy (CDbuf, &play_request, sizeof (play_request));
    device_request ();
    memcpy (&play_request, CDbuf, sizeof (play_request));
    memcpy (&CD_info.error, &play_request.status, sizeof (short));
  }
}

void CD_stop (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
  } stop_request;

  if (CD_installed == CD_INSTALLED)
  {
    if (CD_info.error.busy)
    {
      stop_request.length = sizeof (stop_request);
      stop_request.subunit = 0;
      stop_request.comcode = 133;
      memcpy (CDbuf, &stop_request, sizeof (stop_request));
      device_request ();
      memcpy (&stop_request, CDbuf, sizeof (stop_request));
      memcpy (&CD_info.error, &stop_request.status, sizeof (short));
    }
  }
}


void CD_resume (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
  } stop_request;

  if (CD_installed == CD_INSTALLED)
  {
    if (!CD_info.error.busy)
    {
      stop_request.length = sizeof (stop_request);
      stop_request.subunit = 0;
      stop_request.comcode = 136;
      memcpy (CDbuf, &stop_request, sizeof (stop_request));
      device_request ();
      memcpy (&stop_request, CDbuf, sizeof (stop_request));
      memcpy (&CD_info.error, &stop_request.status, sizeof (short));
    }
  }
}


static void CD_cmd (unsigned char mode)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned char unused[4];
  } tray_request;

  unsigned char CD_mode;

  if (CD_installed == CD_INSTALLED)
  {
    CD_mode = mode;
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 12;
    tray_request.media = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 1;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &CD_mode, 1);
    device_request ();
    memcpy (&tray_request, CDbuf, sizeof (tray_request));
    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  }
}


void CD_getPos ()
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 3;
    tray_request.media = 0;
    tray_request.media = tray_request.sector = tray_request.volid = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 6;
    CD_playInfo.control = 12;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &CD_playInfo, sizeof (TCD_playInfo));
    device_request ();
    memcpy (&tray_request, CDbuf, sizeof (tray_request));
    memcpy (&CD_playInfo, CDbuf2, sizeof (TCD_playInfo));

    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  }
}


static void CD_getDriverName (void)
{
  RMINFO rmi={0};
  TCD_driveList *driveList;
  char *cdstr;

  rmi.x.eax = 0x00001501;
  rmi.x.ebx = DP_OFF (CDbuf);
  rmi.x.es = DP_SEG (CDbuf);

  DPMI_int (0x2F, &rmi);

  driveList = (TCD_driveList *) CDbuf;
  cdstr = (char *) MK_DP (driveList->seg, driveList->off);

  memcpy (CD_driverName, cdstr+10, 8);

  cdstr = strchr (CD_driverName, ' ');
  if (cdstr)
  {
    *cdstr = 0;
  }
}

static void CD_getHandle (void)
{
  int handle;

  CD_handle = 0;
  _dos_open (CD_driverName, O_BINARY | O_RDONLY, &CD_handle);
}


static void CD_ioctl (int command)
{
  RMINFO rmi = {0};

  if (CD_handle)
  {
    *(int *) CDbuf = command;

    rmi.x.eax = 0x4403;
    rmi.x.ebx = CD_handle;
    rmi.x.ecx = 1;
    rmi.x.edx = DP_OFF (CDbuf);
    rmi.x.ds = DP_SEG (CDbuf);

    DPMI_int (0x21, &rmi);
  }
}


void CD_open (void)
{
  if (CD_handle == 0)
  {
    CD_cmd (CMD_EJECT_TRAY);
  }
  else
  {
    CD_ioctl (CMD_EJECT_TRAY);
  }

  while (!CD_info.status.doorOpened)
  {
    CD_getStatus ();
  }

  CD_audioInit ();

  CD_door = CLOSED;
}

void CD_close (void)
{
  if (CD_handle == 0)
  {
    CD_cmd (CMD_CLOSE_TRAY);
  }
  else
  {
    CD_ioctl (CMD_CLOSE_TRAY);
  }

  TIMER_sleep (1);

  CD_getStatus ();

  CD_audioInit ();

  CD_door = OPENED;
}



void CD_reset (void)
{
  if (CD_handle == 0)
  {
    CD_cmd (CMD_RESET);
  }
  else
  {
    CD_ioctl (CMD_RESET);
  }
  CD_audioInit ();
}


void CD_eject (void)
{
  if (!CD_info.status.doorOpened)
  {
    CD_open ();
  }
  else
  {
    switch (CD_door)
    {
      case OPENED:
	CD_open ();
	break;
      case CLOSED:
	CD_close ();
	break;
      default:
      {
	CD_close ();
      }
    }
  }
}



int CD_init (void)
{
  union REGS inregs, outregs;
  struct SREGS sregs;
  memset (&inregs, 0, sizeof inregs);
  memset (&sregs, 0, sizeof sregs);

  if (CD_installed == 0)
  {
    inregs.h.ah = 0x15;
    inregs.h.al = 0x00;
    inregs.x.ebx = 0;
    int386 (0x2f, &inregs, &outregs);
    if (outregs.x.ebx == 0)
    {
      CD_installed = CD_NOTINSTALLED;
    }
    else
    {
      CD_info.drives = outregs.x.ebx;
      CD_info.firstDrive = outregs.x.ecx;

      allocbuffers ();
      if ((CDbuf == NULL) || (CDbuf2 == NULL))
      {
        CD_installed = CD_NOTINSTALLED;
        return CD_NOTINSTALLED;
      }

      CD_installed = CD_INSTALLED;

      CD_getDriverName ();
      CD_getHandle ();

      CD_door = 0;

//      CD_audioInit ();
    }
  }
  return CD_installed;
}


void CD_done (void)
{
  if (CDbuf)
  {
    DPMI_free (CDbuf);
    CDbuf = NULL;
  }
  if (CDbuf2)
  {
    DPMI_free (CDbuf2);
    CDbuf2 = NULL;
  }

  CD_installed = 0;
//  CD_cmd (CLOSE_TRAY);
//  return (CD_info.error.busy == 0);
}


int CD_mediaChange (void)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned short sector;
    unsigned long  volid;
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char media;
  } CD_data;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 3;
    tray_request.media = 0;
    tray_request.media = tray_request.sector = tray_request.volid = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 2;

    CD_data.mode = 0x09;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &CD_data, sizeof (CD_data));
    device_request ();
    memcpy (&tray_request, CDbuf, sizeof (tray_request));
    memcpy (&CD_data, CDbuf2, sizeof (CD_data));

    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
    return CD_data.media;
  }
  return MEDIA_CH_DONTKNOW;
}


static void CD_doorMode (unsigned char doormode)
{
  struct {
    unsigned char length;
    unsigned char subunit;
    unsigned char comcode;
    unsigned short status;
    char ununsed[8];
    unsigned char media;
    unsigned short address_low;
    unsigned short address_high;
    unsigned short bytes;
    unsigned char unused[4];
  } tray_request;

  struct {
    unsigned char mode;
    unsigned char media;
  } CD_data;

  if (CD_installed == CD_INSTALLED)
  {
    tray_request.length = sizeof (tray_request);
    tray_request.subunit = 0;
    tray_request.comcode = 12;
    tray_request.media = 0;
    tray_request.address_low = CDbuf2_low;
    tray_request.address_high = CDbuf2_high;

    tray_request.bytes = 2;
    CD_data.mode = 1;
    CD_data.media = doormode;
    memcpy (CDbuf, &tray_request, sizeof (tray_request));
    memcpy (CDbuf2, &CD_data, sizeof (CD_data));
    device_request ();
    memcpy (&tray_request, CDbuf, sizeof (tray_request));

    memcpy (&CD_info.error, &tray_request.status, sizeof (short));
  }
}

void CD_lock ()
{
//  if (CD_info.status.doorUnlocked)
  {
    CD_doorMode (LOCK);
  }
}

void CD_unlock ()
{
//  if (!CD_info.status.doorUnlocked)
  {
    CD_doorMode (UNLOCK);
  }
}



void CD_playTrack (int tracknum)
{
//  CD_close ();
  if (!CD_info.status.doorOpened)
  {
    CD_stop ();
    if (tracknum != -1)
      CD_setTrack (tracknum);
    CD_play (&CD_info.trackBegin, &CD_info.trackLength);
  }
}

